"""Context: Toolbox || **Category: Fundamental**.

A controller to manage and compile all of the Fundamental models
available in the `toolbox` context. This will be passed as a
`@property` to the `toolbox()` class, giving access to the
Fundamental module and its functions.
"""

from humbldata.core.standard_models.abstract.errors import HumblDataError
from humbldata.core.standard_models.toolbox import ToolboxQueryParams
from humbldata.core.standard_models.toolbox.fundamental.humbl_compass import (
    HumblCompassQueryParams,
)
from humbldata.core.standard_models.toolbox.fundamental.humbl_compass_backtest import (
    HumblCompassBacktestQueryParams,
)
from humbldata.core.utils.env import Env
from humbldata.core.utils.logger import setup_logger

env = Env()
logger = setup_logger("FundamentalController", env.LOGGER_LEVEL)


class Fundamental:
    """
    Module for all Fundamental analysis.

    Attributes
    ----------
    context_params : ToolboxQueryParams
        The standard query parameters for toolbox data.

    Methods
    -------
    humbl_compass(command_params: HumblCompassQueryParams)
        Execute the HumblCompass command.

    """

    def __init__(self, context_params: ToolboxQueryParams):
        self.context_params = context_params

    def humbl_compass_backtest(self, **kwargs: HumblCompassBacktestQueryParams):
        """
        Execute the HumblCompassBacktest command.

        Parameters
        ----------
        **kwargs : HumblCompassBacktestQueryParams
            The command-specific parameters.
        """
        try:
            logger.debug(
                "Initializing HumblCompassBacktest calculation with params: %s",
                kwargs,
            )

            from humbldata.core.standard_models.toolbox.fundamental.humbl_compass_backtest import (
                HumblCompassBacktestFetcher,
            )

            # Instantiate the Fetcher with the query parameters
            fetcher = HumblCompassBacktestFetcher(
                context_params=self.context_params,
                command_params=kwargs,
            )

            logger.debug("Fetching HumblCompassBacktest data")
            return fetcher.fetch_data()

        except Exception as e:
            logger.exception("Error calculating HumblCompassBacktest")
            msg = f"Failed to calculate HumblCompassBacktest: {e!s}"
            raise HumblDataError(msg) from e

    def humbl_compass(self, **kwargs):
        """
        Execute the HumblCompass command.

        Parameters
        ----------
        country : str
            The country or group of countries to analyze
        recommendations : bool, optional
            Whether to include investment recommendations based on the HUMBL regime
        chart : bool, optional
            Whether to return a chart object
        template : str, optional
            The template/theme to use for the plotly figure
        z_score : str, optional
            The time window for z-score calculation

        Returns
        -------
        HumblObject
            The HumblObject containing the transformed data and metadata
        """
        from humbldata.core.standard_models.toolbox.fundamental.humbl_compass import (
            HumblCompassFetcher,
            HumblCompassQueryParams,
        )

        # Convert kwargs to query params and instantiate fetcher.
        command_params = HumblCompassQueryParams(**kwargs)
        fetcher = HumblCompassFetcher(
            context_params=self.context_params, command_params=command_params
        )
        # Get the result which is a HumblObject (with its serialized data),
        # and which also has stored raw transformed data in fetcher._raw_transformed_df.
        result = fetcher.fetch_data()

        # Capture the raw transformed data needed for backtest.
        raw_transformed = fetcher._raw_transformed_df

        # Attach a backtest method to the result.
        def backtest(**backtest_kwargs):
            from humbldata.core.standard_models.toolbox.fundamental.humbl_compass_backtest import (
                HumblCompassBacktestFetcher,
                HumblCompassBacktestQueryParams,
            )

            # Use backtest_kwargs to create backtest command parameters.
            bt_params = HumblCompassBacktestQueryParams(**backtest_kwargs)
            bt_fetcher = HumblCompassBacktestFetcher(
                context_params=self.context_params,
                command_params=bt_params,
                compass_data=raw_transformed,
            )
            return bt_fetcher.fetch_data()

        # Now you can chain: result.backtest(symbols=["SPY"])
        result.backtest = backtest
        return result



import base64
import io
import json
import re
from typing import Any, ClassVar, Generic, TypeVar

import numpy as np
import pandas as pd
import polars as pl
import pyarrow as pa
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    SerializeAsAny,
)

from humbldata.core.standard_models.abstract.chart import Chart
from humbldata.core.standard_models.abstract.errors import HumblDataError
from humbldata.core.standard_models.abstract.query_params import QueryParams
from humbldata.core.standard_models.abstract.tagged import Tagged
from humbldata.core.standard_models.abstract.warnings import Warning_
from humbldata.core.standard_models.portfolio import PortfolioQueryParams
from humbldata.core.standard_models.toolbox import ToolboxQueryParams

T = TypeVar("T")


def extract_subclass_dict(self, attribute_name: str, items: list):
    """
    Extract the dictionary representation of the specified attribute.

    Parameters
    ----------
    attribute_name : str
        The name of the attribute to update in the items list.
    """
    # Check if the attribute exists and has a value
    attribute_value = getattr(self, attribute_name, None)
    if attribute_value:
        # Assuming the attribute has a method called 'model_dump' to get its dictionary representation
        add_item = attribute_value.model_dump()
        add_item_str = str(add_item)
        if len(add_item_str) > 80:
            add_item_str = add_item_str[:80] + "..."
        for i, item in enumerate(items):
            if item.startswith(f"{attribute_name}:"):
                items[i] = f"{attribute_name}: {add_item_str}"
                break

    return items


class HumblObject(Tagged, Generic[T]):
    """HumblObject is the base class for all dta returned from the Toolbox."""

    _user_settings: ClassVar[BaseModel | None] = None
    _system_settings: ClassVar[BaseModel | None] = None

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    results: T | None = Field(
        default=None,
        description="Serializable Logical Plan of the pl.LazyFrame results.",
    )
    equity_data: T | None = Field(
        default=None,
        description="Serialized raw data used in the command calculations.",
    )
    provider: str | None = Field(
        default=None,
        description="Provider name.",
    )
    warnings: list[Warning_] | None = Field(
        default=None,
        description="List of warnings.",
    )
    chart: Chart | list[Chart] | None = Field(
        default=None,
        description="Chart object.",
    )
    extra: dict[str, Any] = Field(
        default_factory=dict,
        description="Extra info.",
    )
    context_params: ToolboxQueryParams | PortfolioQueryParams | None = Field(
        default_factory=ToolboxQueryParams,
        title="Context Parameters",
        description="Context parameters.",
    )
    command_params: SerializeAsAny[QueryParams] | None = Field(
        default=QueryParams,
        title="Command Parameters",
        description="Command-specific parameters.",
    )

    # @field_validator("command_params")
    # def validate_command_params(cls, v):
    #     class_name = v.__class__.__name__
    #     if "QueryParams" in class_name:
    #         return v
    #     msg = "Wrong type for 'command_params', must be subclass of QueryParams"
    #     raise TypeError(msg)

    def __repr__(self) -> str:
        """Human readable representation of the object."""
        items = [
            f"{k}: {v}"[:83] + ("..." if len(f"{k}: {v}") > 83 else "")
            for k, v in self.model_dump().items()
        ]

        # Needed to extract subclass dict correctly
        # items = extract_subclass_dict(self, "command_params", items)

        return f"{self.__class__.__name__}\n\n" + "\n".join(items)

    def to_polars(
        self, collect: bool = True, equity_data: bool = False
    ) -> pl.LazyFrame | pl.DataFrame:
        """
        Deserialize the stored results or return the LazyFrame, and optionally collect them into a Polars DataFrame.

        Parameters
        ----------
        collect : bool, optional
            If True, collects the deserialized LazyFrame into a DataFrame.
            Default is True.
        equity_data : bool, optional
            If True, processes equity_data instead of results.
            Default is False.

        Returns
        -------
        pl.LazyFrame | pl.DataFrame
            The results as a Polars LazyFrame or DataFrame,
            depending on the collect parameter.

        Raises
        ------
        HumblDataError
            If no results or equity data are found to process
        """
        data = self.equity_data if equity_data else self.results

        if data is None:
            raise HumblDataError("No data found.")

        if isinstance(data, pl.LazyFrame):
            out = data
        elif isinstance(data, str):
            with io.StringIO(data) as data_io:
                out = pl.LazyFrame.deserialize(data_io, format="json")
        elif isinstance(data, bytes):
            with io.BytesIO(data) as data_io:
                out = pl.LazyFrame.deserialize(data_io, format="binary")
        else:
            raise HumblDataError(
                "Invalid data type. Expected LazyFrame or serialized string."
            )

        if collect:
            out = out.collect()

        return out

    def to_df(
        self, collect: bool = True, equity_data: bool = False
    ) -> pl.LazyFrame | pl.DataFrame:
        """
        Alias for the `to_polars` method.

        Parameters
        ----------
        collect : bool, optional
            If True, collects the deserialized LazyFrame into a DataFrame.
            Default is True.

        Returns
        -------
        pl.LazyFrame | pl.DataFrame
            The deserialized results as a Polars LazyFrame or DataFrame,
            depending on the collect parameter.
        """
        return self.to_polars(collect=collect, equity_data=equity_data)

    def to_pandas(self, equity_data: bool = False) -> pd.DataFrame:
        """
        Convert the results to a Pandas DataFrame.

        Returns
        -------
        pd.DataFrame
            The results as a Pandas DataFrame.
        """
        return self.to_polars(collect=True, equity_data=equity_data).to_pandas()

    def to_numpy(self, equity_data: bool = False) -> np.ndarray:
        """
        Convert the results to a NumPy array.

        Returns
        -------
        np.ndarray
            The results as a NumPy array.
        """
        return self.to_polars(collect=True, equity_data=equity_data).to_numpy()

    def to_dict(
        self,
        row_wise: bool = False,
        equity_data: bool = False,
        as_series: bool = True,
    ) -> dict | list[dict]:
        """
        Transform the stored data into a dictionary or a list of dictionaries.

        This method allows for the conversion of the internal data
        representation into a more universally accessible format, either
        aggregating the entire dataset into a single dictionary (column-wise)
        or breaking it down into a list of dictionaries, each representing a
        row in the dataset.

        Parameters
        ----------
        row_wise : bool, optional
            Determines the format of the output. If set to True, the method
            returns a list of dictionaries, with each dictionary representing a
            row and its corresponding data as key-value pairs. If set to False,
            the method returns a single dictionary, with column names as keys
            and lists of column data as values. Default is False.

        equity_data : bool, optional
            A flag to specify whether to use equity-specific data for the
            conversion. This parameter allows for flexibility in handling
            different types of data stored within the object. Default is
            False.
        as_series : bool, optional
            If True, the method returns a pl.Series with values as Series. If
            False, the method returns a dict with values as List[Any].
            Default is True.

        Returns
        -------
        dict | list[dict]
            Depending on the `row_wise` parameter, either a dictionary mapping column names to lists of values (if `row_wise` is False) or a list of dictionaries, each representing a row in the dataset (if `row_wise` is True).
        """
        if row_wise:
            return self.to_polars(
                collect=True, equity_data=equity_data
            ).to_dicts()
        return self.to_polars(collect=True, equity_data=equity_data).to_dict(
            as_series=as_series
        )

    def to_arrow(self, equity_data: bool = False) -> pa.Table:
        """
        Convert the results to an Arrow Table.

        Returns
        -------
        pa.Table
            The results as an Arrow Table.
        """
        return self.to_polars(collect=True, equity_data=equity_data).to_arrow()

    def to_struct(
        self, name: str = "results", equity_data: bool = False
    ) -> pl.Series:
        """
        Convert the results to a struct.

        Parameters
        ----------
        name : str, optional
            The name of the struct. Default is "results".

        Returns
        -------
        pl.Struct
            The results as a struct.
        """
        return self.to_polars(collect=True, equity_data=equity_data).to_struct(
            name=name
        )

    def to_json(
        self, equity_data: bool = False, chart: bool = False
    ) -> str | list[str]:
        """
        Convert the results to a JSON string.

        Parameters
        ----------
        equity_data : bool, optional
            A flag to specify whether to use equity-specific data for the
            conversion. Default is False.
        chart : bool, optional
            If True, return all generated charts as a JSON string instead of
            returning the results. Default is False.

        Returns
        -------
        str | list[str]
            The results or charts as a JSON string or list of JSON strings.

        Raises
        ------
        HumblDataError
            If chart is True but no charts are available.
        """
        import json
        from datetime import date, datetime

        from humbldata.core.standard_models.abstract.errors import (
            HumblDataError,
        )

        def json_serial(obj):
            """JSON serializer for objects not serializable by default json code."""
            if isinstance(obj, (datetime, date)):
                return obj.isoformat()
            msg = f"Type {type(obj)} not serializable"
            raise TypeError(msg)

        def decode_base64_numpy_arrays(chart_json: str) -> str:
            """
            Decode base64-encoded numpy arrays in Plotly chart JSON.

            Parameters
            ----------
            chart_json : str
                The JSON string containing the chart data

            Returns
            -------
            str
                The JSON string with decoded numpy arrays
            """
            # Parse the JSON to a Python dict
            if not chart_json:
                return chart_json

            chart_data = json.loads(chart_json)

            # Function to recursively process the chart data
            def process_dict(
                d: dict | list | Any,
            ) -> dict | list | Any:
                if isinstance(d, dict):
                    for key, value in d.items():
                        if (
                            isinstance(value, dict)
                            and "dtype" in value
                            and "bdata" in value
                        ):
                            # This looks like a base64-encoded numpy array
                            try:
                                dtype = value.get("dtype")
                                b64_data = value.get("bdata")

                                if isinstance(b64_data, str):
                                    # Decode the base64 data
                                    binary_data = base64.b64decode(b64_data)

                                    # Convert to numpy array
                                    array = np.frombuffer(
                                        binary_data, dtype=dtype
                                    )

                                    # Replace the encoded data with the actual array values
                                    d[key] = array.tolist()
                            except Exception:
                                # If decoding fails, leave as is
                                pass
                        elif isinstance(value, (dict, list)):
                            process_dict(value)
                elif isinstance(d, list):
                    for i, item in enumerate(d):
                        if isinstance(item, (dict, list)):
                            process_dict(item)
                return d

            # Process the chart data
            processed_data = process_dict(chart_data)

            # Convert back to JSON string with compact formatting (no spaces)
            return json.dumps(processed_data, separators=(",", ":"))

        if chart:
            if self.chart is None:
                msg = f"You set `.to_json(chart=True)` but there were no charts. Make sure `chart=True` in {self.command_params.__class__.__name__}"
                raise HumblDataError(msg)

            if isinstance(self.chart, list):
                return [
                    decode_base64_numpy_arrays(chart.content)
                    if chart and chart.content
                    else ""
                    for chart in self.chart
                    if chart
                ]
            else:
                return (
                    decode_base64_numpy_arrays(self.chart.content)
                    if self.chart.content
                    else ""
                )
        else:
            data = self.to_polars(
                collect=True, equity_data=equity_data
            ).to_dict(as_series=False)
            return json.dumps(data, default=json_serial, separators=(",", ":"))

    def is_empty(self, equity_data: bool = False) -> bool:
        """
        Check if the results are empty.

        Returns
        -------
        bool
            True if the results are empty, False otherwise.
        """
        return self.to_polars(collect=True, equity_data=equity_data).is_empty()

    def show(self) -> None:
        """Show the chart."""
        if isinstance(self.chart, list):
            for chart in self.chart:
                if chart and chart.fig:
                    chart.fig.show()
                else:
                    msg = "Chart object is missing or incomplete."
                    raise HumblDataError(msg)
        elif not self.chart or not self.chart.fig:
            msg = "Chart not found."
            raise HumblDataError(msg)



import plotly.graph_objs as go
import polars as pl

from humbldata.core.standard_models.abstract.chart import Chart, ChartTemplate
from humbldata.core.utils import plotly_theme  # noqa: F401


def create_backtest_plot(
    equity_data: pl.LazyFrame,
    regime_date_summary: pl.LazyFrame,
    symbol: str = "",
    template: ChartTemplate = ChartTemplate.plotly,
) -> go.Figure:
    """
    Generate a backtest plot with equity data and regime overlay shading.

    Parameters
    ----------
    equity_data : pl.LazyFrame
        The LazyFrame containing date and close price data.
    regime_date_summary : pl.LazyFrame
        The LazyFrame containing regime data with humbl_regime, regime_instance_id,
        start_date, and end_date columns.
    symbol : str, optional
        The symbol for which the backtest plot is to be generated.
    template : ChartTemplate, optional
        The template to be used for styling the plot.

    Returns
    -------
    go.Figure
        A plotly figure object representing the backtest data with regime shading.
    """
    # Collect data from LazyFrames
    equity_df = equity_data.collect()
    regime_df = regime_date_summary.collect()

    # Create the figure
    fig = go.Figure()

    # Add equity price trace
    fig.add_trace(
        go.Scatter(
            x=equity_df.select("date").to_series(),
            y=equity_df.select("close").to_series(),
            name="Price",
            line=dict(color="black", width=2),
        )
    )

    # Define colors for each regime
    regime_colors = {
        "humblBOOM": "rgba(0, 255, 0, 0.4)",  # Green with transparency
        "humblBOUNCE": "rgba(0, 0, 255, 0.4)",  # Blue with transparency
        "humblBLOAT": "rgba(255, 165, 0, 0.4)",  # Orange with transparency
        "humblBUST": "rgba(255, 0, 0, 0.4)",  # Red with transparency
    }

    # Add shaded regions for each regime period
    for row in regime_df.iter_rows(named=True):
        regime = row["humbl_regime"]
        start_date = row["start_date"]
        end_date = row["end_date"]
        color = regime_colors.get(
            regime, "rgba(128, 128, 128, 0.2)"
        )  # Default gray if regime not found

        # Add a vertical rectangle for each regime period
        fig.add_vrect(
            x0=start_date,
            x1=end_date,
            fillcolor=color,
            opacity=1,
            layer="below",
            line_width=0,
        )

    # Update layout
    title = (
        f"Humbl Compass Backtest for {symbol}"
        if symbol
        else "Humbl Compass Backtest"
    )
    fig.update_layout(
        title=title,
        xaxis_title="Date",
        yaxis_title="Price",
        template=template,
        legend_title="Legend",
    )

    return fig


def generate_plot_for_symbol(
    equity_data: pl.LazyFrame,
    regime_date_summary: pl.LazyFrame,
    symbol: str,
    template: ChartTemplate = ChartTemplate.plotly,
) -> Chart:
    """
    Generate a backtest plot for a specific symbol.

    Parameters
    ----------
    equity_data : pl.LazyFrame
        The LazyFrame containing equity data for all symbols.
    regime_date_summary : pl.LazyFrame
        The LazyFrame containing regime data for all symbols.
    symbol : str
        The symbol for which to generate the plot.
    template : ChartTemplate, optional
        The template to be used for styling the plot.

    Returns
    -------
    Chart
        A Chart object containing the generated backtest plot for the specified symbol.
    """
    # Filter data for the specific symbol
    filtered_equity_data = equity_data.filter(pl.col("symbol") == symbol)
    filtered_regime_data = regime_date_summary.filter(
        pl.col("symbol") == symbol
    )

    # Create the plot
    fig = create_backtest_plot(
        filtered_equity_data, filtered_regime_data, symbol, template
    )

    return Chart(content=fig.to_json(), fig=fig)


def generate_plots(
    equity_data: pl.LazyFrame,
    regime_date_summary: pl.LazyFrame,
    template: ChartTemplate = ChartTemplate.plotly,
) -> list[Chart]:
    """
    Context: Toolbox || Category: Fundamental || Subcategory: Humbl Compass Backtest || **Command: generate_plots()**.

    Generate backtest plots for each unique symbol in the given dataframes.

    Parameters
    ----------
    equity_data : pl.LazyFrame
        The LazyFrame containing equity data for all symbols.
    regime_date_summary : pl.LazyFrame
        The LazyFrame containing regime data for all symbols.
    template : ChartTemplate, optional
        The template to be used for styling the plots.

    Returns
    -------
    list[Chart]
        A list of Chart objects, each representing a plot for a unique symbol.
    """
    # Get unique symbols
    symbols = equity_data.select("symbol").unique().collect().to_series()

    # Generate plots for each symbol
    plots = [
        generate_plot_for_symbol(
            equity_data, regime_date_summary, symbol, template
        )
        for symbol in symbols
    ]

    return plots



import plotly
import plotly.graph_objs as go
import polars as pl

from humbldata.core.standard_models.abstract.chart import Chart, ChartTemplate
from humbldata.core.standard_models.abstract.humblobject import HumblObject
from humbldata.core.utils import plotly_theme  # noqa: F401

# Color constants for easy modification
MOMENTUM_COLORS = {
    "positive": "blue",
    "negative": "yellow",
    "default": "blue",  # Used when no momentum data
}

CHANNEL_COLORS = {
    "top": "red",
    "bottom": "green",
}


def create_historical_plot(
    data: pl.DataFrame,
    symbol: str,
    template: ChartTemplate = ChartTemplate.plotly,
) -> go.Figure:
    """
    Generate a historical plot for a given symbol from the provided data.

    Parameters
    ----------
    data : pl.DataFrame
        The dataframe containing historical data including dates, bottom prices, close prices, and top prices.
    symbol : str
        The symbol for which the historical plot is to be generated.
    template : ChartTemplate
        The template to be used for styling the plot.

    Returns
    -------
    go.Figure
        A plotly figure object representing the historical data of the given symbol.
    """
    filtered_data = data.filter(pl.col("symbol") == symbol)

    fig = go.Figure()

    # Check if momentum_signal exists
    has_momentum = "momentum_signal" in filtered_data.schema.names()

    if has_momentum:
        # Get all series at once
        dates = filtered_data.select("date").to_series()
        prices = filtered_data.select("recent_price").to_series()
        signals = filtered_data.select("momentum_signal").to_series()

        # Split data into positive and negative momentum
        pos_dates = []
        pos_prices = []
        neg_dates = []
        neg_prices = []

        # Add first point to appropriate list
        if signals[0] == 1:
            pos_dates.append(dates[0])
            pos_prices.append(prices[0])
        else:
            neg_dates.append(dates[0])
            neg_prices.append(prices[0])

        # Process each point
        for i in range(len(signals)):
            if signals[i] == 1:
                pos_dates.append(dates[i])
                pos_prices.append(prices[i])
                # If next point exists and has different momentum, add None for gap
                if i < len(signals) - 1 and signals[i + 1] != 1:
                    pos_dates.append(dates[i + 1])
                    pos_prices.append(prices[i + 1])
                    pos_dates.append(None)
                    pos_prices.append(None)
            else:
                neg_dates.append(dates[i])
                neg_prices.append(prices[i])
                # If next point exists and has different momentum, add None for gap
                if i < len(signals) - 1 and signals[i + 1] != 0:
                    neg_dates.append(dates[i + 1])
                    neg_prices.append(prices[i + 1])
                    neg_dates.append(None)
                    neg_prices.append(None)

        # Add positive momentum trace
        if pos_dates:
            fig.add_trace(
                go.Scatter(
                    x=pos_dates,
                    y=pos_prices,
                    mode="lines",
                    line=dict(color=MOMENTUM_COLORS["positive"], width=2),
                    name="Price (Positive Momentum)",
                )
            )

        # Add negative momentum trace
        if neg_dates:
            fig.add_trace(
                go.Scatter(
                    x=neg_dates,
                    y=neg_prices,
                    mode="lines",
                    line=dict(color=MOMENTUM_COLORS["negative"], width=2),
                    name="Price (Negative Momentum)",
                )
            )
    else:
        # Original price trace if no momentum
        fig.add_trace(
            go.Scatter(
                x=filtered_data.select("date").to_series(),
                y=filtered_data.select("recent_price").to_series(),
                name="Recent Price",
                line=dict(color=MOMENTUM_COLORS["default"]),
            )
        )

    fig.add_trace(
        go.Scatter(
            x=filtered_data.select("date").to_series(),
            y=filtered_data.select("bottom_price").to_series(),
            name="Bottom Price",
            line=dict(color=CHANNEL_COLORS["bottom"]),
        )
    )
    fig.add_trace(
        go.Scatter(
            x=filtered_data.select("date").to_series(),
            y=filtered_data.select("top_price").to_series(),
            name="Top Price",
            line=dict(color=CHANNEL_COLORS["top"]),
        )
    )
    fig.update_layout(
        title=f"Historical Mandelbrot Channel for {symbol}",
        xaxis_title="Date",
        yaxis_title="Price",
        template=template,
    )
    return fig


def create_current_plot(
    data: pl.DataFrame,
    equity_data: pl.DataFrame,
    symbol: str,
    template: ChartTemplate = ChartTemplate.plotly,
) -> go.Figure:
    """
    Generate a current plot for a given symbol from the provided data and equity data.

    Parameters
    ----------
    data : pl.DataFrame
        The dataframe containing historical data including top and bottom prices.
    equity_data : pl.DataFrame
        The dataframe containing current equity data including dates and close prices.
    symbol : str
        The symbol for which the current plot is to be generated.
    template : ChartTemplate
        The template to be used for styling the plot.

    Returns
    -------
    go.Figure
        A plotly figure object representing the current data of the given symbol.
    """
    filtered_data = data.filter(pl.col("symbol") == symbol)
    equity_data = equity_data.filter(pl.col("symbol") == symbol)

    fig = go.Figure()

    # Check if momentum_signal exists
    has_momentum = "momentum_signal" in equity_data.schema.names()

    if has_momentum:
        # Get all series at once
        dates = equity_data.select("date").to_series()
        prices = equity_data.select("close").to_series()
        signals = equity_data.select("momentum_signal").to_series()

        # Split data into positive and negative momentum
        pos_dates = []
        pos_prices = []
        neg_dates = []
        neg_prices = []

        # Add first point to appropriate list
        if signals[0] == 1:
            pos_dates.append(dates[0])
            pos_prices.append(prices[0])
        else:
            neg_dates.append(dates[0])
            neg_prices.append(prices[0])

        # Process each point
        for i in range(len(signals)):
            if signals[i] == 1:
                pos_dates.append(dates[i])
                pos_prices.append(prices[i])
                # If next point exists and has different momentum, add None for gap
                if i < len(signals) - 1 and signals[i + 1] != 1:
                    pos_dates.append(dates[i + 1])
                    pos_prices.append(prices[i + 1])
                    pos_dates.append(None)
                    pos_prices.append(None)
            else:
                neg_dates.append(dates[i])
                neg_prices.append(prices[i])
                # If next point exists and has different momentum, add None for gap
                if i < len(signals) - 1 and signals[i + 1] != 0:
                    neg_dates.append(dates[i + 1])
                    neg_prices.append(prices[i + 1])
                    neg_dates.append(None)
                    neg_prices.append(None)

        # Add positive momentum trace
        if pos_dates:
            fig.add_trace(
                go.Scatter(
                    x=pos_dates,
                    y=pos_prices,
                    mode="lines",
                    line=dict(color=MOMENTUM_COLORS["positive"], width=2),
                    name="Price (Positive Momentum)",
                )
            )

        # Add negative momentum trace
        if neg_dates:
            fig.add_trace(
                go.Scatter(
                    x=neg_dates,
                    y=neg_prices,
                    mode="lines",
                    line=dict(color=MOMENTUM_COLORS["negative"], width=2),
                    name="Price (Negative Momentum)",
                )
            )
    else:
        # Original price trace if no momentum
        fig.add_trace(
            go.Scatter(
                x=equity_data.select("date").to_series(),
                y=equity_data.select("close").to_series(),
                name="Recent Price",
                line=dict(color=MOMENTUM_COLORS["default"]),
            )
        )

    fig.add_hline(
        y=filtered_data.select("top_price").row(0)[0],
        line=dict(color=CHANNEL_COLORS["top"], width=2),
        name="Top Price",
    )
    fig.add_hline(
        y=filtered_data.select("bottom_price").row(0)[0],
        line=dict(color=CHANNEL_COLORS["bottom"], width=2),
        name="Bottom Price",
    )
    fig.update_layout(
        title=f"Current Mandelbrot Channel for {symbol}",
        xaxis_title="Date",
        yaxis_title="Price",
        template=template,
    )
    return fig


def is_historical_data(data: pl.DataFrame) -> bool:
    """
    Check if the provided dataframe contains historical data based on the uniqueness of dates.

    Parameters
    ----------
    data : pl.DataFrame
        The dataframe to check for historical data presence.

    Returns
    -------
    bool
        Returns True if the dataframe contains historical data (more than one unique date), otherwise False.
    """
    return data.select("date").to_series().unique().shape[0] > 1


def generate_plot_for_symbol(
    data: pl.DataFrame,
    equity_data: pl.DataFrame,
    symbol: str,
    template: ChartTemplate = ChartTemplate.plotly,
    momentum_data: pl.DataFrame | None = None,
) -> Chart:
    """
    Generate a plot for a specific symbol that is filtered from the original DF.

    This function will check if the data provided is a Historical or Current
    Mandelbrot Channel data. If it is historical, it will generate a historical
    plot. If it is current, it will generate a current plot.

    Parameters
    ----------
    data : pl.DataFrame
        The dataframe containing Mandelbrot channel data for all symbols.
    equity_data : pl.DataFrame
        The dataframe containing equity data for all symbols.
    symbol : str
        The symbol for which to generate the plot.
    template : ChartTemplate
        The template/theme to use for the plotly figure.
    momentum_data : pl.DataFrame | None, optional
        Optional dataframe containing historical momentum signals. If provided,
        will be used to color the equity price trace in current plot mode.

    Returns
    -------
    Chart
        A Chart object containing the generated plot for the specified symbol.
    """
    if is_historical_data(data):
        out = create_historical_plot(data, symbol, template)
    else:
        # If we have momentum data, merge it with equity data
        if momentum_data is not None:
            equity_data = equity_data.join(
                momentum_data.select(["date", "symbol", "momentum_signal"]),
                on=["date", "symbol"],
                how="left",
            )
        out = create_current_plot(data, equity_data, symbol, template)

    return Chart(content=out.to_json(), fig=out)


def generate_plots(
    data: pl.LazyFrame,
    equity_data: pl.LazyFrame,
    template: ChartTemplate = ChartTemplate.plotly,
    momentum_data: pl.LazyFrame | None = None,
) -> list[Chart]:
    """
    Context: Toolbox || Category: Technical || Subcategory: Mandelbrot Channel || **Command: generate_plots()**.

    Generate plots for each unique symbol in the given dataframes.

    Parameters
    ----------
    data : pl.LazyFrame
        The LazyFrame containing the symbols and HumblChannelData
    equity_data : pl.LazyFrame
        The LazyFrame containing equity data for the symbols.
    template : ChartTemplate
        The template/theme to use for the plotly figure.
    momentum_data : pl.LazyFrame | None, optional
        Optional LazyFrame containing historical momentum signals. If provided,
        will be used to color the equity price trace in current plot mode.

    Returns
    -------
    list[Chart]
        A list of Chart objects, each representing a plot for a unique symbol.
    """
    symbols = data.select("symbol").unique().collect().to_series()

    # Collect DataFrames once for efficiency
    data_df = data.collect()
    equity_df = equity_data.collect()
    momentum_df = momentum_data.collect() if momentum_data is not None else None

    plots = [
        generate_plot_for_symbol(
            data_df,
            equity_df,
            symbol,
            template,
            momentum_df,
        )
        for symbol in symbols
    ]
    return plots



"""
**Context: Toolbox || Category: Technical || Command: momentum**.

The momentum Command Module. This module calculates various types of momentum/Rate of Change (ROC)
indicators for time series data.
"""

from typing import Literal

import polars as pl

from humbldata.core.standard_models.abstract.errors import HumblDataError
from humbldata.core.utils.env import Env
from humbldata.core.utils.logger import setup_logger
from humbldata.toolbox.toolbox_helpers import (
    _check_required_columns,
    _set_sort_cols,
    _window_format,
)

env = Env()
logger = setup_logger("humbl_momentum.model", level=env.LOGGER_LEVEL)


def _calc_log_roc(data: pl.LazyFrame, window_days: int = 1) -> pl.LazyFrame:
    """Calculate logarithmic rate of change."""
    try:
        # First calculate the momentum column
        data = data.with_columns(
            (pl.col("close").log() - pl.col("close").shift(window_days).log())
            .over("symbol")
            .alias("momentum")
        )

        # Then calculate the signal using the momentum column
        return data.with_columns(
            (pl.col("momentum") > 0)
            .over("symbol")
            .cast(pl.Int8)
            .alias("momentum_signal")
        )
    except Exception as e:
        logger.error(f"Error calculating log ROC: {str(e)}")
        raise HumblDataError("Failed to calculate logarithmic ROC") from e


def _calc_simple_roc(data: pl.LazyFrame, window_days: int = 1) -> pl.LazyFrame:
    """Calculate simple rate of change."""
    try:
        # First calculate the momentum column
        data = data.with_columns(
            (
                (pl.col("close") - pl.col("close").shift(window_days))
                / pl.col("close").shift(window_days)
            )
            .over("symbol")
            .alias("momentum")
        )

        # Then calculate the signal using the momentum column
        return data.with_columns(
            (pl.col("momentum") > 0)
            .over("symbol")
            .cast(pl.Int8)
            .alias("momentum_signal")
        )
    except Exception as e:
        logger.error(f"Error calculating simple ROC: {str(e)}")
        raise HumblDataError("Failed to calculate simple ROC") from e


def _calc_shift(data: pl.LazyFrame, window_days: int = 1) -> pl.LazyFrame:
    """Calculate simple time series shift."""
    try:
        # First add the shifted column
        data = data.with_columns(
            pl.col("close").shift(window_days).over("symbol").alias("shifted")
        )

        # Then calculate the signal using the shifted column
        return data.with_columns(
            (pl.col("close") > pl.col("shifted"))
            .over("symbol")
            .cast(pl.Int8)
            .alias("momentum_signal")
        )
    except Exception as e:
        logger.error(f"Error calculating shift: {str(e)}")
        raise HumblDataError("Failed to calculate shift") from e


def calc_humbl_momentum(
    data: pl.DataFrame | pl.LazyFrame,
    method: Literal["log", "simple", "shift"] = "log",
    window: str = "1d",
) -> pl.LazyFrame:
    """
    Context: Toolbox || Category: Technical || Command: calc_humbl_momentum.

    Calculate momentum/Rate of Change (ROC) for time series data.

    Parameters
    ----------
    data : pl.DataFrame | pl.LazyFrame
        Input data containing at minimum 'close' and 'date' columns
    method : Literal["log", "simple", "shift"], default "log"
        Method to calculate momentum:
        - "log": Logarithmic ROC
        - "simple": Simple ROC (percentage change)
        - "shift": Simple time series shift with binary signal
    window : str, default "1d"
        Window to calculate momentum over

    Returns
    -------
    pl.LazyFrame
        DataFrame with added columns based on method:
        - log/simple: adds 'momentum' column
        - shift: adds 'shifted' and 'momentum' columns (1 when price > shifted price, 0 otherwise)

    Raises
    ------
    HumblDataError
        If required columns are missing or calculation fails
    """
    try:
        # Validate input
        required_cols = ["close", "date"]
        _check_required_columns(data, *required_cols)
        sort_cols = _set_sort_cols(data, "symbol", "date")

        window_days: int = _window_format(
            window, _return_timedelta=True, _avg_trading_days=False
        ).days

        logger.debug(
            f"Calculating momentum using method: {method}, window: {window_days}"
        )

        # Convert to LazyFrame if DataFrame
        data = data.lazy() if isinstance(data, pl.DataFrame) else data

        # Sort by date to ensure proper shifting
        data = data.sort("date")

        # Calculate momentum based on method
        if method == "log":
            logger.debug("Using logarithmic ROC calculation")
            result = _calc_log_roc(data, window_days)
        elif method == "shift":
            logger.debug("Using simple shift calculation")
            result = _calc_shift(data, window_days)
        else:
            logger.debug("Using simple ROC calculation")
            result = _calc_simple_roc(data, window_days)

        return result.sort(sort_cols)

    except Exception as e:
        logger.error(f"Failed to calculate momentum: {str(e)}")
        raise HumblDataError(f"Momentum calculation failed: {str(e)}") from e



from datetime import datetime
from typing import Any, ClassVar

import polars as pl
from lumibot.backtesting import YahooDataBacktesting
from lumibot.brokers import Alpaca
from lumibot.entities.position import Position
from lumibot.strategies.strategy import Strategy
from lumibot.traders import Trader

from humbldata.core.utils.env import Env
from humbldata.toolbox.toolbox_controller import Toolbox


class HumblChannelSingleLong(Strategy):
    """Strategy that trades based on HumblChannel indicators."""

    parameters: ClassVar[dict[str, Any]] = {
        "symbol": "SPY",
        "quantity": 1,
        "buy_threshold_pct": 0.005,  # 2% threshold
        "sell_threshold_pct": 0.001,  # 2% threshold
    }

    def initialize(self):
        self.sleeptime = "1D"

    def before_market_opens(self):
        current_date = self.get_round_day()
        # current_date = self.get_datetime().strftime("%Y-%m-%d")
        # self.log_message(f"Current Trading Day: {current_date}", color="blue")

        # Calculates the humblCHANNEL from 2000-01-01 to current date, returns only one observation
        self.vars.daily_humbl_channel = (
            Toolbox(
                symbols=self.parameters["symbol"],
                interval="1d",
                start_date="2000-01-01",
                end_date=current_date,
                membership="admin",
            )
            .technical.humbl_channel(
                historical=False,
                window="1mo",
                _boundary_group_down=False,
                chart=False,
                live_price=False,
            )
            .to_polars()
        )

        # self.log_message(
        #     f"HumblChannel Bottom Price: {self.vars.daily_humbl_channel.select('bottom_price')}",
        #     color="blue",
        # )
        # self.log_message(
        #     f"HumblChannel Top Price: {self.vars.daily_humbl_channel.select('top_price')}",
        #     color="blue",
        # )

    def on_trading_iteration(self):
        symbol = self.parameters["symbol"]
        cash = self.get_cash()

        # Get current price and position
        current_price = self.get_last_price(symbol)
        position = self.get_position(symbol)

        # Get bottom and top prices from HumblChannel
        bottom_price = self.vars.daily_humbl_channel.select("bottom_price").row(
            0
        )[0]
        top_price = self.vars.daily_humbl_channel.select("top_price").row(0)[0]

        # Calculate price differences as percentages
        bottom_diff_pct = abs(current_price - bottom_price) / bottom_price
        top_diff_pct = abs(current_price - top_price) / top_price

        # First check sell condition - if we have a position and price is near top
        if (
            position is not None
            and abs(top_diff_pct) <= self.parameters["sell_threshold_pct"]
        ):
            # Sell all shares
            order = self.create_order(symbol, position.quantity, "sell")
            self.submit_order(order)
            return  # Exit to prevent buy on same day

        # Then check buy condition - if we have no position and price is near bottom
        if (
            position is None
            and abs(bottom_diff_pct) <= self.parameters["buy_threshold_pct"]
        ):
            # Calculate maximum shares we can buy with available cash
            max_shares = int(cash / current_price)
            if max_shares > 0:
                order = self.create_order(symbol, max_shares, "buy")
                self.submit_order(order)


if __name__ == "__main__":
    is_live = False

    if is_live:
        env = Env()
        ALPACA_CONFIG = {
            "API_KEY": env.ALPACA_API_KEY,
            "API_SECRET": env.ALPACA_API_SECRET,
            "PAPER": True,
        }

        trader = Trader()
        broker = Alpaca(ALPACA_CONFIG)

        strategy = HumblChannelSingleLong(
            broker=broker,
            parameters={
                "symbol": "SPY",
                "quantity": 1,
            },
        )

        trader.add_strategy(strategy)
        strategy_executors = trader.run_all()

    else:
        # Run backtest
        backtesting_start = datetime(2010, 1, 1)
        backtesting_end = datetime(2024, 12, 10)

        results = HumblChannelSingleLong.backtest(
            YahooDataBacktesting,
            backtesting_start,
            backtesting_end,
            parameters={
                "symbol": "SPY",
            },
            benchmark_asset="SPY",
        )



"""
HumblCompass Standard Model.

Context: Toolbox || Category: Fundamental || Command: humbl_compass.

This module is used to define the QueryParams and Data model for the
HumblCompass command.
"""

import warnings
from datetime import datetime
from enum import Enum
from typing import Literal, TypeVar

import pandera.polars as pa
import polars as pl
from openbb import obb
from pydantic import BaseModel, Field

from humbldata.core.standard_models.abstract.chart import ChartTemplate
from humbldata.core.standard_models.abstract.data import Data
from humbldata.core.standard_models.abstract.humblobject import HumblObject
from humbldata.core.standard_models.abstract.query_params import QueryParams
from humbldata.core.standard_models.abstract.warnings import (
    HumblDataWarning,
    collect_warnings,
)
from humbldata.core.standard_models.toolbox import ToolboxQueryParams
from humbldata.core.utils.env import Env
from humbldata.core.utils.logger import log_start_end, setup_logger
from humbldata.toolbox.fundamental.humbl_compass.view import generate_plots
from humbldata.toolbox.toolbox_helpers import (
    _window_format,
    _window_format_monthly,
)

env = Env()
Q = TypeVar("Q", bound=ToolboxQueryParams)
logger = setup_logger("HumblCompassFetcher", level=env.LOGGER_LEVEL)

HUMBLCOMPASS_QUERY_DESCRIPTIONS = {
    "example_field1": "Description for example field 1",
    "example_field2": "Description for example field 2",
    "chart": "Whether to return a chart object.",
    "template": "The template/theme to use for the plotly figure.",
}

HUMBLCOMPASS_DATA_DESCRIPTIONS = {
    "date": "The date of the data point",
    "country": "The country or group of countries the data represents",
    "cpi": "Consumer Price Index (CPI) value",
    "cpi_3m_delta": "3-month delta of CPI",
    "cpi_1yr_zscore": "1-year rolling z-score of CPI",
    "cli": "Composite Leading Indicator (CLI) value",
    "cli_3m_delta": "3-month delta of CLI",
    "cli_1yr_zscore": "1-year rolling z-score of CLI",
    "humbl_regime": "HUMBL Regime classification based on CPI and CLI values",
}


class AssetRecommendation(str, Enum):
    """Asset recommendation categories."""

    EQUITIES = "Equities"
    CREDIT = "Credit"
    COMMODITIES = "Commodities"
    FX = "FX"
    FIXED_INCOME = "Fixed Income"
    USD = "USD"
    GOLD = "Gold"
    TECHNOLOGY = "Technology"
    CONSUMER_DISCRETIONARY = "Consumer Discretionary"
    MATERIALS = "Materials"
    INDUSTRIALS = "Industrials"
    UTILITIES = "Utilities"
    REITS = "REITs"
    CONSUMER_STAPLES = "Consumer Staples"
    FINANCIALS = "Financials"
    ENERGY = "Energy"
    HEALTH_CARE = "Health Care"
    TELECOM = "Telecom"
    HIGH_BETA = "High Beta"
    MOMENTUM = "Momentum"
    CYCLICALS = "Cyclicals"
    SECULAR_GROWTH = "Secular Growth"
    LOW_BETA = "Low Beta"
    DEFENSIVES = "Defensives"
    VALUE = "Value"
    DIVIDEND_YIELD = "Dividend Yield"
    QUALITY = "Quality"
    CYCLICAL_GROWTH = "Cyclical Growth"
    SMALL_CAPS = "Small Caps"
    MID_CAPS = "Mid Caps"
    BDCS = "BDCs"
    CONVERTIBLES = "Convertibles"
    HY_CREDIT = "HY Credit"
    EM_DEBT = "EM Debt"
    TIPS = "TIPS"
    SHORT_DURATION_TREASURIES = "Short Duration Treasuries"
    MORTGAGE_BACKED_SECURITIES = "Mortgage Backed Securities"
    MEDIUM_DURATION_TREASURIES = "Medium Duration Treasuries"
    LONG_DURATION_TREASURIES = "Long Duration Treasuries"
    IG_CREDIT = "Investment Grade Credit"
    MUNIS = "Municipal Bonds"
    PREFERREDS = "Preferreds"
    EM_LOCAL_CURRENCY = "Emerging Market Local Currency"
    LEVERAGED_LOANS = "Leveraged Loans"


class RecommendationCategory(BaseModel):
    """Category-specific recommendations with rationale."""

    best: list[AssetRecommendation]
    worst: list[AssetRecommendation]
    rationale: str


class RegimeRecommendations(BaseModel):
    """Complete set of recommendations for a specific regime."""

    asset_classes: RecommendationCategory
    equity_sectors: RecommendationCategory
    equity_factors: RecommendationCategory
    fixed_income: RecommendationCategory
    regime_description: str
    key_risks: list[str]
    last_updated: datetime = Field(default_factory=datetime.utcnow)


class HumblCompassQueryParams(QueryParams):
    """
    QueryParams model for the HumblCompass command, a Pydantic v2 model.

    Parameters
    ----------
    country : Literal
        The country or group of countries to collect humblCOMPASS data for.
    cli_start_date : str
        The adjusted start date for CLI data collection.
    cpi_start_date : str
        The adjusted start date for CPI data collection.
    z_score : Optional[str]
        The time window for z-score calculation (e.g., "1 year", "18 months").
    chart : bool
        Whether to return a chart object.
    template : Literal
        The template/theme to use for the plotly figure.
    """

    country: Literal[
        "g20",
        "g7",
        "asia5",
        "north_america",
        "europe4",
        "australia",
        "brazil",
        "canada",
        "china",
        "france",
        "germany",
        "india",
        "indonesia",
        "italy",
        "japan",
        "mexico",
        "south_africa",
        "south_korea",
        "spain",
        "turkey",
        "united_kingdom",
        "united_states",
        "all",
    ] = Field(
        default="united_states",
        title="Country for humblCOMPASS data",
        description=HUMBLCOMPASS_QUERY_DESCRIPTIONS.get("country", ""),
    )
    cli_start_date: str = Field(
        default=None,
        title="Adjusted start date for CLI data",
        description="The adjusted start date for CLI data collection.",
    )
    cpi_start_date: str = Field(
        default=None,
        title="Adjusted start date for CPI data",
        description="The adjusted start date for CPI data collection.",
    )
    z_score: str | None = Field(
        default=None,
        title="Z-score calculation window",
        description="The time window for z-score calculation (e.g., '1 year', '18 months').",
    )
    chart: bool = Field(
        default=False,
        title="Results Chart",
        description=HUMBLCOMPASS_QUERY_DESCRIPTIONS.get("chart", ""),
    )
    template: Literal[
        "humbl_dark",
        "humbl_light",
        "ggplot2",
        "seaborn",
        "simple_white",
        "plotly",
        "plotly_white",
        "plotly_dark",
        "presentation",
        "xgridoff",
        "ygridoff",
        "gridon",
        "none",
    ] = Field(
        default="humbl_dark",
        title="Plotly Template",
        description=HUMBLCOMPASS_QUERY_DESCRIPTIONS.get("template", ""),
    )
    recommendations: bool = Field(
        default=False,
        title="Investment Recommendations",
        description="Whether to include investment recommendations based on the HUMBL regime.",
    )


REGIME_RECOMMENDATIONS: dict[str, RegimeRecommendations] = {
    "humblBOOM": RegimeRecommendations(
        asset_classes=RecommendationCategory(
            best=[
                AssetRecommendation.EQUITIES,
                AssetRecommendation.CREDIT,
                AssetRecommendation.COMMODITIES,
                AssetRecommendation.FX,
            ],
            worst=[AssetRecommendation.FIXED_INCOME, AssetRecommendation.USD],
            rationale="Strong growth and inflation expectations favor risk assets",
        ),
        equity_sectors=RecommendationCategory(
            best=[
                AssetRecommendation.TECHNOLOGY,
                AssetRecommendation.CONSUMER_DISCRETIONARY,
                AssetRecommendation.MATERIALS,
                AssetRecommendation.INDUSTRIALS,
            ],
            worst=[
                AssetRecommendation.UTILITIES,
                AssetRecommendation.REITS,
                AssetRecommendation.CONSUMER_STAPLES,
                AssetRecommendation.FINANCIALS,
            ],
            rationale="Growth sectors outperform in expansionary environments",
        ),
        equity_factors=RecommendationCategory(
            best=[
                AssetRecommendation.HIGH_BETA,
                AssetRecommendation.MOMENTUM,
                AssetRecommendation.CYCLICALS,
                AssetRecommendation.SECULAR_GROWTH,
            ],
            worst=[
                AssetRecommendation.LOW_BETA,
                AssetRecommendation.DEFENSIVES,
                AssetRecommendation.VALUE,
                AssetRecommendation.DIVIDEND_YIELD,
            ],
            rationale="Risk-on factors perform well in growth environments",
        ),
        fixed_income=RecommendationCategory(
            best=[
                AssetRecommendation.BDCS,
                AssetRecommendation.CONVERTIBLES,
                AssetRecommendation.HY_CREDIT,
                AssetRecommendation.EM_DEBT,
            ],
            worst=[
                AssetRecommendation.TIPS,
                AssetRecommendation.SHORT_DURATION_TREASURIES,
                AssetRecommendation.MORTGAGE_BACKED_SECURITIES,
                AssetRecommendation.MEDIUM_DURATION_TREASURIES,
            ],
            rationale="Credit risk outperforms duration risk",
        ),
        regime_description="Accelerating growth and decelerating inflation favors risk assets with big market multiples in equities, junk bonds, and real growth. This is the time to BUY 'anything'!",
        key_risks=[
            "Inflation Resurgence",
            "Policy Tightening",
            "Valuation Expansion",
        ],
    ),
    "humblBUST": RegimeRecommendations(
        asset_classes=RecommendationCategory(
            best=[
                AssetRecommendation.FIXED_INCOME,
                AssetRecommendation.GOLD,
                AssetRecommendation.USD,
            ],
            worst=[
                AssetRecommendation.COMMODITIES,
                AssetRecommendation.EQUITIES,
                AssetRecommendation.CREDIT,
                AssetRecommendation.FX,
            ],
            rationale="Deflationary environment favors safe-haven assets and cash",
        ),
        equity_sectors=RecommendationCategory(
            best=[
                AssetRecommendation.CONSUMER_STAPLES,
                AssetRecommendation.UTILITIES,
                AssetRecommendation.REITS,
                AssetRecommendation.HEALTH_CARE,
            ],
            worst=[
                AssetRecommendation.ENERGY,
                AssetRecommendation.TECHNOLOGY,
                AssetRecommendation.INDUSTRIALS,
                AssetRecommendation.FINANCIALS,
            ],
            rationale="Defensive sectors outperform in contractionary environments",
        ),
        equity_factors=RecommendationCategory(
            best=[
                AssetRecommendation.LOW_BETA,
                AssetRecommendation.DIVIDEND_YIELD,
                AssetRecommendation.QUALITY,
                AssetRecommendation.DEFENSIVES,
            ],
            worst=[
                AssetRecommendation.HIGH_BETA,
                AssetRecommendation.MOMENTUM,
                AssetRecommendation.CYCLICALS,
                AssetRecommendation.SECULAR_GROWTH,
            ],
            rationale="Low-risk factors outperform in risk-off environments",
        ),
        fixed_income=RecommendationCategory(
            best=[
                AssetRecommendation.LONG_DURATION_TREASURIES,
                AssetRecommendation.MEDIUM_DURATION_TREASURIES,
                AssetRecommendation.IG_CREDIT,
                AssetRecommendation.MUNIS,
            ],
            worst=[
                AssetRecommendation.PREFERREDS,
                AssetRecommendation.EM_LOCAL_CURRENCY,
                AssetRecommendation.BDCS,
                AssetRecommendation.LEVERAGED_LOANS,
            ],
            rationale="Duration risk outperforms credit risk in deflationary environments",
        ),
        regime_description="Deflationary environment favors treasuries and cash while avoiding high yield credit and stocks",
        key_risks=[
            "Policy response delay",
            "Deflation spiral",
            "Credit market stress",
        ],
    ),
    "humblBOUNCE": RegimeRecommendations(
        asset_classes=RecommendationCategory(
            best=[
                AssetRecommendation.COMMODITIES,
                AssetRecommendation.EQUITIES,
                AssetRecommendation.CREDIT,
                AssetRecommendation.FX,
            ],
            worst=[
                AssetRecommendation.FIXED_INCOME,
                AssetRecommendation.USD,
            ],
            rationale="Rising yields and improving growth favor risk assets",
        ),
        equity_sectors=RecommendationCategory(
            best=[
                AssetRecommendation.TECHNOLOGY,
                AssetRecommendation.CONSUMER_DISCRETIONARY,
                AssetRecommendation.INDUSTRIALS,
                AssetRecommendation.MATERIALS,
            ],
            worst=[
                AssetRecommendation.TELECOM,
                AssetRecommendation.UTILITIES,
                AssetRecommendation.REITS,
                AssetRecommendation.CONSUMER_STAPLES,
            ],
            rationale="Growth and cyclical sectors benefit from improving conditions",
        ),
        equity_factors=RecommendationCategory(
            best=[
                AssetRecommendation.SECULAR_GROWTH,
                AssetRecommendation.MOMENTUM,
                AssetRecommendation.CYCLICAL_GROWTH,
                AssetRecommendation.SMALL_CAPS,
            ],
            worst=[
                AssetRecommendation.LOW_BETA,
                AssetRecommendation.VALUE,
                AssetRecommendation.DIVIDEND_YIELD,
                AssetRecommendation.DEFENSIVES,
            ],
            rationale="Growth and high-beta factors lead in recovery phases",
        ),
        fixed_income=RecommendationCategory(
            best=[
                AssetRecommendation.CONVERTIBLES,
                AssetRecommendation.BDCS,
                AssetRecommendation.PREFERREDS,
                AssetRecommendation.LEVERAGED_LOANS,
            ],
            worst=[
                AssetRecommendation.LONG_DURATION_TREASURIES,
                AssetRecommendation.MEDIUM_DURATION_TREASURIES,
                AssetRecommendation.MUNIS,
                AssetRecommendation.IG_CREDIT,
            ],
            rationale="Credit-sensitive sectors outperform as rates rise and spreads tighten",
        ),
        regime_description="Recovery phase with rising bond yields, improving financials, and strengthening commodities",
        key_risks=[
            "False recovery",
            "Policy tightening too soon",
            "Inflation resurgence",
        ],
    ),
    "humblBLOAT": RegimeRecommendations(
        asset_classes=RecommendationCategory(
            best=[
                AssetRecommendation.GOLD,
                AssetRecommendation.COMMODITIES,
            ],
            worst=[
                AssetRecommendation.CREDIT,
            ],
            rationale="USD devaluation and money printing favor real assets",
        ),
        equity_sectors=RecommendationCategory(
            best=[
                AssetRecommendation.UTILITIES,
                AssetRecommendation.TECHNOLOGY,
                AssetRecommendation.ENERGY,
                AssetRecommendation.INDUSTRIALS,
            ],
            worst=[
                AssetRecommendation.FINANCIALS,
                AssetRecommendation.REITS,
                AssetRecommendation.MATERIALS,
                AssetRecommendation.TELECOM,
            ],
            rationale="Sectors with pricing power and real asset exposure outperform",
        ),
        equity_factors=RecommendationCategory(
            best=[
                AssetRecommendation.SECULAR_GROWTH,
                AssetRecommendation.MOMENTUM,
                AssetRecommendation.MID_CAPS,
                AssetRecommendation.LOW_BETA,
            ],
            worst=[
                AssetRecommendation.SMALL_CAPS,
                AssetRecommendation.DIVIDEND_YIELD,
                AssetRecommendation.VALUE,
                AssetRecommendation.DEFENSIVES,
            ],
            rationale="Quality growth outperforms as real growth slows but inflation accelerates",
        ),
        fixed_income=RecommendationCategory(
            best=[
                AssetRecommendation.MUNIS,
                AssetRecommendation.EM_DEBT,
                AssetRecommendation.LONG_DURATION_TREASURIES,
                AssetRecommendation.TIPS,
            ],
            worst=[
                AssetRecommendation.BDCS,
                AssetRecommendation.PREFERREDS,
                AssetRecommendation.CONVERTIBLES,
                AssetRecommendation.LEVERAGED_LOANS,
            ],
            rationale="High-quality duration outperforms as real yields decline",
        ),
        regime_description="Central Bank Policy response to economic slowdown creates illusion of growth through inflation acceleration while real growth slows",
        key_risks=[
            "Stagflation",
            "Policy Credibility Loss",
            "Real Growth Deterioration",
        ],
    ),
}


class HumblCompassData(Data):
    """
    Data model for the humbl_compass command, a Pandera.Polars Model.

    This Data model is used to validate data in the `.transform_data()` method of the `HumblCompassFetcher` class.
    """

    date_month_start: pl.Date = pa.Field(
        default=None,
        title="Date",
        description=HUMBLCOMPASS_DATA_DESCRIPTIONS["date"],
    )
    country: pl.Utf8 = pa.Field(
        default=None,
        title="Country",
        description=HUMBLCOMPASS_DATA_DESCRIPTIONS["country"],
    )
    cpi: pl.Float64 = pa.Field(
        default=None,
        title="Consumer Price Index (CPI)",
        description=HUMBLCOMPASS_DATA_DESCRIPTIONS["cpi"],
    )
    cpi_3m_delta: pl.Float64 = pa.Field(
        default=None,
        title="Consumer Price Index (CPI) 3-Month Delta",
        description=HUMBLCOMPASS_DATA_DESCRIPTIONS["cpi_3m_delta"],
    )
    cpi_zscore: pl.Float64 | None = pa.Field(
        default=None,
        title="Consumer Price Index (CPI) 1-Year Z-Score",
        description=HUMBLCOMPASS_DATA_DESCRIPTIONS["cpi_1yr_zscore"],
        nullable=True,
    )
    cli: pl.Float64 = pa.Field(
        default=None,
        title="Composite Leading Indicator (CLI)",
        description=HUMBLCOMPASS_DATA_DESCRIPTIONS["cli"],
    )
    cli_3m_delta: pl.Float64 = pa.Field(
        default=None,
        title="Composite Leading Indicator (CLI) 3-Month Delta",
        description=HUMBLCOMPASS_DATA_DESCRIPTIONS["cli_3m_delta"],
    )
    cli_zscore: pl.Float64 | None = pa.Field(
        default=None,
        title="Composite Leading Indicator (CLI) 1-Year Z-Score",
        description=HUMBLCOMPASS_DATA_DESCRIPTIONS["cli_1yr_zscore"],
        nullable=True,
    )
    humbl_regime: pl.Utf8 = pa.Field(
        default=None,
        title="HUMBL Regime",
        description=HUMBLCOMPASS_DATA_DESCRIPTIONS["humbl_regime"],
    )


class HumblCompassFetcher:
    """
    Fetcher for the HumblCompass command.

    Parameters
    ----------
    context_params : ToolboxQueryParams
        The context parameters for the Toolbox query.
    command_params : HumblCompassQueryParams
        The command-specific parameters for the HumblCompass query.

    Attributes
    ----------
    context_params : ToolboxQueryParams
        Stores the context parameters passed during initialization.
    command_params : HumblCompassQueryParams
        Stores the command-specific parameters passed during initialization.
    data : pl.DataFrame
        The raw data extracted from the data provider, before transformation.

    Methods
    -------
    transform_query()
        Transform the command-specific parameters into a query.
    extract_data()
        Extracts the data from the provider and returns it as a Polars DataFrame.
    transform_data()
        Transforms the command-specific data according to the HumblCompass logic.
    fetch_data()
        Execute TET Pattern.
    backtest(**kwargs)
        Execute backtest calculations on the compass data using additional parameters.

    Returns
    -------
    HumblObject
        results : HumblCompassData
            Serializable results.
        provider : Literal['fmp', 'intrinio', 'polygon', 'tiingo', 'yfinance']
            Provider name.
        warnings : Optional[List[Warning_]]
            List of warnings.
        chart : Optional[Chart]
            Chart object.
        context_params : ToolboxQueryParams
            Context-specific parameters.
        command_params : HumblCompassQueryParams
            Command-specific parameters.
    """

    def __init__(
        self,
        context_params: ToolboxQueryParams,
        command_params: HumblCompassQueryParams,
    ):
        """
        Initialize the HumblCompassFetcher with context and command parameters.

        Parameters
        ----------
        context_params : ToolboxQueryParams
            The context parameters for the Toolbox query.
        command_params : HumblCompassQueryParams
            The command-specific parameters for the HumblCompass query.
        """
        self.context_params = context_params
        self.command_params = command_params
        self.warnings = []  # Initialize warnings list

    def transform_query(self):
        """
        Transform the command-specific parameters into a query.

        If command_params is not provided, it initializes a default HumblCompassQueryParams object.
        Calculates adjusted start dates for CLI and CPI data collection.
        """
        if not self.command_params:
            self.command_params = HumblCompassQueryParams()
        elif isinstance(self.command_params, dict):
            self.command_params = HumblCompassQueryParams(**self.command_params)

        # Calculate adjusted start dates
        if isinstance(self.context_params.start_date, str):
            start_date = pl.Series(
                [datetime.strptime(self.context_params.start_date, "%Y-%m-%d")]
            )
        else:
            start_date = pl.Series([self.context_params.start_date])

        # Calculate z-score window in months
        self.z_score_months = 0
        if (
            self.command_params.z_score is not None
            and self.context_params.membership != "humblPEON"
        ):
            z_score_months_str = _window_format(
                self.command_params.z_score, _return_timedelta=False
            )
            self.z_score_months = _window_format_monthly(z_score_months_str)
        if 0 < self.z_score_months < 3:
            logger.warning(
                "Z-score calculation requires a minimum of 3 months. Setting z_score to 3 months."
            )
            self.z_score_months = 3
        elif self.context_params.membership == "humblPEON":
            logger.warning(
                "Z-score is not calculated for humblPEON membership level."
            )

        cli_start_date = start_date.dt.offset_by(
            f"-{4 + self.z_score_months}mo"
        ).dt.strftime("%Y-%m-%d")[0]
        cpi_start_date = start_date.dt.offset_by(
            f"-{4 + self.z_score_months}mo"
        ).dt.strftime("%Y-%m-%d")[0]

        # Update the command_params with the new start dates
        self.command_params = self.command_params.model_copy(
            update={
                "cli_start_date": cli_start_date,
                "cpi_start_date": cpi_start_date,
            }
        )

        logger.info(
            f"CLI start date: {self.command_params.cli_start_date} and CPI start date: {self.command_params.cpi_start_date}. "
            f"Dates are adjusted to account for CLI data release lag and z-score calculation window."
        )

    def extract_data(self):
        """
        Extract the data from the provider and returns it as a Polars DataFrame.

        Returns
        -------
        self
            The HumblCompassFetcher instance with extracted data.
        """
        # Collect CLI Data
        self.oecd_cli_data = (
            obb.economy.composite_leading_indicator(
                start_date=self.command_params.cli_start_date,
                end_date=self.context_params.end_date,
                provider="oecd",
                country=self.command_params.country,
            )
            .to_polars()
            .lazy()
            .rename({"value": "cli"})
            .with_columns(
                [pl.col("date").dt.month_start().alias("date_month_start")]
            )
        )

        # Collect YoY CPI Data
        self.oecd_cpi_data = (
            obb.economy.cpi(
                start_date=self.command_params.cpi_start_date,
                end_date=self.context_params.end_date,
                frequency="monthly",
                country=self.command_params.country,
                transform="yoy",
                provider="oecd",
                harmonized=False,
                expenditure="total",
            )
            .to_polars()
            .lazy()
            .rename({"value": "cpi"})
            .with_columns(
                [pl.col("date").dt.month_start().alias("date_month_start")]
            )
        )
        return self

    @collect_warnings
    def transform_data(self):
        """
        Transform the command-specific data according to the humbl_compass logic.

        Returns
        -------
        self
            The HumblCompassFetcher instance with transformed data.
        """
        # Combine CLI and CPI data
        # CLI data is released before CPI data, so we use a left join
        combined_data = (
            self.oecd_cli_data.join(
                self.oecd_cpi_data.with_columns(
                    [
                        pl.col("country")
                        .str.replace("_", " ")
                        .str.to_titlecase(),
                        pl.col("date_month_start").cast(pl.Date),
                    ]
                ),
                on=["date_month_start", "country"],
                how="inner",  # Changed to inner from left to only return rows where dates & country match
                suffix="_cpi",
            )
            .sort("date_month_start")
            .with_columns(
                [
                    pl.col("country").cast(pl.Utf8),
                    pl.col("cli").cast(pl.Float64),
                    pl.col("cpi").cast(pl.Float64) * 100,
                ]
            )
            .rename({"date": "date_cli"})
            .select(
                [
                    "date_month_start",
                    "date_cli",
                    "date_cpi",
                    "country",
                    "cli",
                    "cpi",
                ]
            )
        )

        # Calculate 3-month deltas with adaptive shift based on data frequency
        transformed_data = combined_data.with_columns(
            [
                # Calculate the average time difference between observations in days
                (
                    pl.col("date_month_start")
                    .diff()
                    .dt.total_days()  # Get total days directly
                    .mean()
                    .over("country")
                    .alias("avg_days_between_obs")
                )
            ]
        )

        # Now use this to determine the appropriate shift
        transformed_data = transformed_data.with_columns(
            [
                pl.when(pl.col("avg_days_between_obs") >= 85)  # ~3 months
                .then(
                    pl.col("cli") - pl.col("cli").shift(1)
                )  # For quarterly data
                .otherwise(
                    pl.col("cli") - pl.col("cli").shift(3)
                )  # For monthly data
                .alias("cli_3m_delta"),
                pl.when(pl.col("avg_days_between_obs") >= 85)  # ~3 months
                .then(
                    pl.col("cpi") - pl.col("cpi").shift(1)
                )  # For quarterly data
                .otherwise(
                    pl.col("cpi") - pl.col("cpi").shift(3)
                )  # For monthly data
                .alias("cpi_3m_delta"),
            ]
        ).drop("avg_days_between_obs")

        # Add this after calculating 3-month deltas in transform_data()
        transformed_data = transformed_data.with_columns(
            [
                pl.when(
                    (pl.col("cpi_3m_delta") > 0) & (pl.col("cli_3m_delta") < 0)
                )
                .then(pl.lit("humblBLOAT"))
                .when(
                    (pl.col("cpi_3m_delta") > 0) & (pl.col("cli_3m_delta") > 0)
                )
                .then(pl.lit("humblBOUNCE"))
                .when(
                    (pl.col("cpi_3m_delta") < 0) & (pl.col("cli_3m_delta") > 0)
                )
                .then(pl.lit("humblBOOM"))
                .when(
                    (pl.col("cpi_3m_delta") < 0) & (pl.col("cli_3m_delta") < 0)
                )
                .then(pl.lit("humblBUST"))
                .otherwise(None)
                .alias("humbl_regime")
            ]
        )

        # Calculate z-scores only if self.z_score_months is greater than 0 and membership is not humblPEON
        if (
            self.z_score_months >= 3
            and self.context_params.membership != "humblPEON"
        ):
            transformed_data = transformed_data.with_columns(
                [
                    pl.when(
                        pl.col("cli").count().over("country")
                        >= self.z_score_months
                    )
                    .then(
                        (
                            pl.col("cli")
                            - pl.col("cli").rolling_mean(self.z_score_months)
                        )
                        / pl.col("cli").rolling_std(self.z_score_months)
                    )
                    .alias("cli_zscore"),
                    pl.when(
                        pl.col("cpi").count().over("country")
                        >= self.z_score_months
                    )
                    .then(
                        (
                            pl.col("cpi")
                            - pl.col("cpi").rolling_mean(self.z_score_months)
                        )
                        / pl.col("cpi").rolling_std(self.z_score_months)
                    )
                    .alias("cpi_zscore"),
                ]
            )

        # Select columns based on whether z-scores were calculated
        columns_to_select = [
            pl.col("date_month_start"),
            pl.col("country"),
            pl.col("cpi").round(2),
            pl.col("cpi_3m_delta").round(2),
            pl.col("cli").round(2),
            pl.col("cli_3m_delta").round(2),
            pl.col("humbl_regime"),
        ]

        if (
            self.z_score_months >= 3
            and self.context_params.membership != "humblPEON"
        ):
            columns_to_select.extend(
                [
                    pl.col("cpi_zscore").round(2),
                    pl.col("cli_zscore").round(2),
                ]
            )

        transformed_data = transformed_data.select(columns_to_select)

        # Validate the data using HumblCompassData and collect result BEFORE serialization.
        raw_data = (
            transformed_data.collect().drop_nulls()
        )  # <-- keep raw data for backtest
        self._raw_transformed_df = raw_data  # store raw transformed data

        self.transformed_data = HumblCompassData(raw_data).lazy()

        # Generate chart if requested
        self.chart = None
        if self.command_params.chart:
            self.chart = generate_plots(
                self.transformed_data,
                template=ChartTemplate(self.command_params.template),
            )

        # Add warning if z_score is None
        if self.command_params.z_score is None:
            warnings.warn(
                "Z-score defaulted to None. No z-score data will be calculated.",
                category=HumblDataWarning,
                stacklevel=1,
            )

        # Add recommendations if requested
        if self.command_params.recommendations:
            latest_regime = (
                self.transformed_data.select(pl.col("humbl_regime"))
                .collect()
                .row(-1)[0]
            )

            if latest_regime not in REGIME_RECOMMENDATIONS:
                warnings.warn(
                    f"No recommendations available for regime: {latest_regime}",
                    category=HumblDataWarning,
                    stacklevel=1,
                )
            else:
                recommendations = REGIME_RECOMMENDATIONS[latest_regime]
                if not hasattr(self, "extra"):
                    self.extra = {}
                self.extra["humbl_regime_recommendations"] = (
                    recommendations.model_dump()
                )

        # Finally, serialize final transformed data.
        self.transformed_data = self.transformed_data.serialize(format="binary")
        return self

    @log_start_end(logger=logger)
    def fetch_data(self):
        """
        Execute TET Pattern.

        This method executes the query transformation, data fetching and
        transformation process by first calling `transform_query` to prepare the query parameters, then
        extracting the raw data using `extract_data` method, and finally
        transforming the raw data using `transform_data` method.

        Returns
        -------
        HumblObject
            The HumblObject containing the transformed data and metadata.
        """
        logger.debug("Running .transform_query()")
        self.transform_query()
        logger.debug("Running .extract_data()")
        self.extract_data()
        logger.debug("Running .transform_data()")
        self.transform_data()

        # Initialize warnings list if it doesn't exist
        if not hasattr(self.context_params, "warnings"):
            self.context_params.warnings = []

        # Combine warnings from both sources
        all_warnings = self.context_params.warnings + self.warnings

        return HumblObject(
            results=self.transformed_data,
            provider=self.context_params.provider,
            warnings=all_warnings,  # Use combined warnings
            chart=self.chart,
            context_params=self.context_params,
            command_params=self.command_params,
            extra=self.extra
            if hasattr(self, "extra")
            else {},  # pipe in extra from transform_data()
        )



import contextlib
import functools
from typing import Any, Callable, List, Optional, Type, Union
from warnings import WarningMessage, catch_warnings

from pydantic import BaseModel


class Warning_(BaseModel):  # noqa: N801, D101
    category: str
    message: str


def cast_warning(w: WarningMessage) -> Warning_:  # noqa: D103
    return Warning_(
        category=w.category.__name__,
        message=str(w.message),
    )


class HumblDataWarning(Warning):  # noqa: D101
    """Base warning class for HumblData warnings."""

    pass


# Create a function to convert a Warning to a Warning_ object
def warning_to_warning_(warning_obj: Warning) -> Warning_:
    """Convert a Warning object to a Warning_ object."""
    return Warning_(
        category=warning_obj.__class__.__name__,
        message=str(warning_obj),
    )


class WarningCollector:
    """Context manager and decorator for collecting warnings."""

    def __init__(self):
        self.warnings: list[Warning_] = []

    def __enter__(self):
        self._recorded = []
        self._showwarning_orig = None
        self._catch_ctx = catch_warnings(record=True)
        self._warnings = self._catch_ctx.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._catch_ctx.__exit__(exc_type, exc_val, exc_tb)
        self.warnings.extend([cast_warning(w) for w in self._warnings])
        return False

    def collect_warnings(self, func: Callable) -> Callable:
        """
        Collect warnings from a function.

        Parameters
        ----------
        func : Callable
            The function to wrap

        Returns
        -------
        Callable
            The wrapped function that collects warnings
        """

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with self:
                result = func(*args, **kwargs)
            return result

        return wrapper


# Global collector for use in decorators
warning_collector = WarningCollector()


def collect_warnings(func: Callable) -> Callable:
    """
    Collect warnings from a function and store them in the global collector.

    Parameters
    ----------
    func : Callable
        The function to wrap

    Returns
    -------
    Callable
        The wrapped function that collects warnings
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with warning_collector:
            result = func(*args, **kwargs)

        # If this is a method of a class that has a warnings attribute, add the warnings
        if args and hasattr(args[0], "warnings"):
            if not isinstance(args[0].warnings, list):
                args[0].warnings = []

            # Only add warnings that don't already exist in the list
            for warning in warning_collector.warnings:
                if not any(
                    w.message == warning.message for w in args[0].warnings
                ):
                    args[0].warnings.append(warning)

        return result

    return wrapper


@contextlib.contextmanager
def collect_warnings_context():
    """
    Context manager to collect warnings.

    Yields
    ------
    List[Warning_]
        The list of collected warnings
    """
    collector = WarningCollector()
    with collector:
        yield collector.warnings


# Helper function to create a Warning_ object directly
def create_warning(category: str, message: str) -> Warning_:
    """
    Create a Warning_ object directly.

    Parameters
    ----------
    category : str
        The category of the warning
    message : str
        The warning message

    Returns
    -------
    Warning_
        The created Warning_ object
    """
    return Warning_(category=category, message=message)
